/*
   Frustum Culling Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to perform frustum culling in a 3D scene.  Using frustum
   culling will speed up your frame rate by a great deal.  What is frustum
   culling you might ask?  Well it is the process of eliminating geometry from the rendering
   pipeline of objects you can not see anyhow.  In other words if the user can not see something,
   or expample a character, then we will not waste resources rendering it.  How does this work?
   Well in this demo we are displaying 1000 cubes in the scene.  We take the bounding box
   of each object and test it against the frustum.  If the bouding box is within the frustum then
   we will draw the cube, else we will not.  Since we are dealing with cubes we don't need to
   calculate the bounding geometery (that is done in the bounding box demos in case you want
   to know how), all we need in fact is the position of the cube and the size of the cube and
   we are in business.  I created a class called CFrustum that we will use to do all of this work.

   A frustum is made up of 6 sides.  And top, bottom, left, right, far, and near plane.  What
   we are doing is creating a frustum around all we can see as far as we can see it.  Then when
   we render our objects we test if the object is within the frustum.  If so we draw it else we
   dont.  Since we are using cubes (drawing 1000) it only makes sense that we test the bounding
   box of each cube against the frustum.  Since a cube itself can be used as the bounding box
   this makes our job easy.

   Controls:

      Escape Key  - Exit the appliction.
      Space bar   - Used to switch between using the frustum culling and not.
*/


// Demo: Frustum Culling.


#define WIN32_LEAN_AND_MEAN   // Trims down the libraries.
#define VC_LEANMEAN           // Trims even further.


#include<windows.h>           // Windows header file.
#include<gl/gl.h>             // OpenGL header.
#include<gl/glu.h>            // OpenGL utility.
#include<gl/glaux.h>          // OpenGL aux.
#include"CFrustum.h"          // Frustum class.
#include"CVector.h"           // CVector4 class.
#include"CGLFont.h"           // OpenGL font class.

#define MAX_CUBES 1000        // Total amount of cubes we have in our scene.


// Function prototypes...
bool InitializeGL();          // Initialize OpenGL.
void RenderScene();           // Renders each frame.
int GetFPS();                 // Returns the frames per second.
void Shutdown();              // Clears out all resources used by this demo.

HDC g_HDC;                    // Global HDC used in the RenderScene() to swap buffers.


CGLFont font;                 // Font object used to display text to the screen.
CFrustum Frustum;             // Frustum object used to determine if a object is visible or not.

CVector4 CubePos[MAX_CUBES];  // List of all the 1000 cubes positions.

float speed = 0.01f;          // Speed the cubes are moving along the y axis.
float CubeSize = 0.8f;        // Size of the cubes.
bool RenderAll = false;       // True if we are to ignore frustum culling and display all cubes.


void SetupPixelFormat(HDC hDC)
{
   int nPixelFormat;

   static PIXELFORMATDESCRIPTOR pfd = {
         sizeof(PIXELFORMATDESCRIPTOR),   // size of structure.
         1,                               // always 1.
         PFD_DRAW_TO_WINDOW |             // support window
         PFD_SUPPORT_OPENGL |             // support OpenGl
         PFD_DOUBLEBUFFER,                // support double buffering
         PFD_TYPE_RGBA,                   // support RGBA
         16,                              // bit color mode
         0, 0, 0, 0, 0, 0,                // ignore color bits
         0,                               // no alpha buffer
         0,                               // ignore shift bit
         0,                               // no accumulation buffer
         0, 0, 0, 0,                      // ignore accumulation bits.
         16,                              // number of depth buffer bits.
         0,                               // number of stencil buffer bits.
         0,                               // 0 means no auxiliary buffer
         PFD_MAIN_PLANE,                  // The main drawing plane
         0,                               // this is reserved
         0, 0, 0 };                       // layer masks ignored.

   // this chooses the best pixel format and returns index.
   nPixelFormat = ChoosePixelFormat(hDC, &pfd);

   // This set pixel format to device context.
   SetPixelFormat(hDC, nPixelFormat, &pfd);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static HGLRC hRC;
   static HDC hDC;
   int width, height;


   switch(message)
      {
         case WM_CREATE:                     // Windows creation.
            hDC = GetDC(hwnd);               // This gets the device context for our window.
            g_HDC = hDC;                     // Assigns the global device context to this one.
            SetupPixelFormat(hDC);           // Call the pixel format function.

            hRC = wglCreateContext(hDC);     // Creates the rendering context.
            wglMakeCurrent(hDC, hRC);        // Makes the rendering context.
            return 0;
            break;


         case WM_CLOSE:                      // Close message.
         case WM_DESTROY:
            wglMakeCurrent(hDC, NULL);
            wglDeleteContext(hRC);           // Deletes the rendering context.

            PostQuitMessage(0);              // Says close the program.
            return 0;
            break;


         case WM_SIZE:                       // re-size message.
            height = HIWORD(lParam);         // This gets the height of the window.
            width = LOWORD(lParam);          // This gets the width of the window.

            if(height==0)                    // we don't want it to be possible for a
               {                             // height of 0.  If it is 0 me make it 1.
                  height = 1;
               }

            glViewport(0, 0, width, height); // resets the viewport to new dimensions.
            glMatrixMode(GL_PROJECTION);     // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            // calculate the aspect ratio of the window.
            gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 1000.0f);

            glMatrixMode(GL_MODELVIEW);      // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            font.SetWidthHeight(width, height);

            return 0;
            break;


         case WM_KEYDOWN:
		      switch(wParam)
               { 
                  // Close the app if the user presses escape.
	               case VK_ESCAPE:
                     PostQuitMessage(0);
                     break;

                  case VK_SPACE:
                     RenderAll = !RenderAll;
                     break;

                  default:
                     break;
               }
            break;

         default: // Always have a default just in case.
            break;
      }
 
   // What this does is pass all of the unhandled messages to DefWindowProc.
   return (DefWindowProc(hwnd, message, wParam, lParam));
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;
   WNDCLASSEX windowClass;
   HWND hwnd;
   bool isFinished;

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);        // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;    // style of the window.
   windowClass.lpfnWndProc = WndProc;              // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                     // Extra class information.
   windowClass.cbWndExtra = 0;                     // Extra window information.
   windowClass.hInstance = hInstance;              // Handle of application Instance.
   // Handle of application Icon.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);// mouse cursor
   windowClass.hbrBackground = NULL;               // background color.
   windowClass.lpszMenuName = NULL;                // name of the main menu.
   windowClass.lpszClassName = "UltimateGameProgrammingClass";// window class name.
   // icon when minimized.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass))
      return 0;

   // Create window...
   hwnd = CreateWindowEx(NULL,// The extended window style.
                        "UltimateGameProgrammingClass",// window Class name.
                        "Frustum Culling Demo - By the Programming Ace.",// window name.
                        WS_OVERLAPPEDWINDOW | WS_VISIBLE |// The window style.
                        WS_SYSMENU | WS_CLIPCHILDREN |// window style.
                        WS_CLIPSIBLINGS,// window style.
                        100, 100,// window x, y coordinate.
                        640,
                        480,// window width and height.
                        NULL,// handle to parent window.
                        NULL,// menu.
                        hInstance,// handle to app instance.
                        NULL);   // pointer to window creation data.
   
   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW); // This shows the window.
   UpdateWindow(hwnd);        // This forces a paint message.

   isFinished = false;        // False = running, True = not running.


   // If initialize fail (return false), then we don't want the program to run.
   if(!InitializeGL()) isFinished = true;

   // This is the messsage loop.
   while(!isFinished)
      {
         if(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     isFinished = true;
                  }

               TranslateMessage(&msg);  // Translate any messages.
               DispatchMessage(&msg);   // Dispatch any messages.
            }
         else
            {
               RenderScene();          // Render a frame.
            }
      }

   Shutdown();

   return (int)msg.wParam;
}


bool InitializeGL()
{
   glClearColor(0.0f, 0.0f, 0.0f, 1.0f);  // Clear the screen to a set color.
   glEnable(GL_LIGHTING);                 // Enable lighting.
   glEnable(GL_DEPTH_TEST);					// Enable depth testing.
   glShadeModel(GL_SMOOTH);               // Smooth shading in our scenes.
    
   // Lighting colors and position.
   float ambientLight[] = {0.2f, 0.2f, 0.2f, 1.0f};
   float diffuseLight[] = {1.0, 1.0, 1.0, 1.0}; 
   float specularLight[] = {1.0, 1.0, 1.0, 1.0};
   float lightPos[] = {2.0f, 6.0f, 0.0f, 1.0f};

   // Set the light properties.
   glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight);
   glLightfv(GL_LIGHT0, GL_SPECULAR, specularLight);
   glLightfv(GL_LIGHT0, GL_AMBIENT, ambientLight);
   glLightfv(GL_LIGHT0, GL_POSITION, lightPos);

   // Enable the first light.
   glEnable(GL_LIGHT0);

   // Next we will initialize the positions of all the cubes.  This part is not important
   // but if you would like to know how it works I'll explain.  There are 1000 cubes being
   // displayed in 10 rows and 10 columns 10 times.  So each layer has 100 cubes in it.
   // This next loop will start initializing at the upper left corner.  It will move down 10
   // and if y == 12.6 (which is 9 x 1.4 (0 - 9 is ten numbers)) then it will reset 0 and
   // move the x 1.7 position over.  It will do this 10 times until there are 100 cubes, 10
   // going across and 10 going down.  Then it will reset the y and the x after that and add
   // 1 to the z.  This will allow it to create a whole new set of 100 cubes.  This continues
   // until 1000 cubes have been initialized.  Hey I needed some way to initialize 1000 cube
   // positions and I was not going to do it the long way.

   float x = 0.0f; float y = 0.0f; float z = 0.0f;

   // Initialize the cubes position.
   for(int i = 0; i < MAX_CUBES; i++)
      {
         CubePos[i] = CVector4(-8.0f + x, -2.0f - y, -2.0f - z);

         y += 1.4f;              // Move down a little for the next cube.

         // If we moved down 10 times (1.4 x 9 is 12.6 (9 because 0 - 9 is 10 numbers)).
         if(y >= 12.6f)
            {
               y = 0;            // Reset y;
               x += 1.7f;        // Move over 1.7th of a position.
               
               // If we across down 10 times (1.7 x 9 is 15.3 (9 because 0 - 9 is 10 numbers)).
               if(x >= 15.3f)
                  {
                     x = 0;      // Reset x.
                     z += 1.0f;  // Move back one to start a new set of cubes.
                  }
            }
      }


   // Initialize the font object so we can display output using it.
   if(!font.Initialize(g_HDC, 20, "Arial"))
      return false;

   // If all went well we return true.
   return true;
}


void RenderScene()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();

   int cubes = 0;             // Stores the total number of cubes rendered (for output purposes).
   char text[64] = {0};       // Stores some text to be displayed to the screen.
   float model[16] = {0};     // Array to store the modelview matrix.
   float proj[16] = {0};      // Array to store the projection matrix.

   // Grab the model view matrix and the projection matrix and store them in our arrays.
   glGetFloatv(GL_PROJECTION_MATRIX, proj);
   glGetFloatv(GL_MODELVIEW_MATRIX, model);

   // Now we calculate the viewing frustum.  In a game situation you would only want to do
   // this is the camera has moved in some way.
   Frustum.CalculateFrustum(model, proj);

   // Set the color to white.
	glColor3f(1.0f, 1.0f ,1.0f);

   // Loop through all of the cubes and draw the ones that are visible.
   for(int i = 0; i < MAX_CUBES; i++)
      {
         glPushMatrix();
            
            // Only draw is we want them all drawn (RenderAll == true) or if the cube
            // is visible in the frustum.
	         if(RenderAll || Frustum.isBoxVisiable(CubePos[i].x, CubePos[i].y,
                                                  CubePos[i].z, CubeSize))
               {
                  // Move this cube to its set position.
                  glTranslatef(CubePos[i].x, CubePos[i].y, CubePos[i].z);

                  // Draw a cube then increment the total cubes that have been rendered.
                  auxSolidCube(CubeSize);
                  cubes++;
               }

            // We are scrolling upwards so increase the y axis by the amount in speed.
            CubePos[i].y += speed;

            // Error checking...we want to keep seeing these boxes so once it reaches a
            // certain height we reset the y axis for this certain box.
            if(CubePos[i].y > 5.0f)
               CubePos[i].y = -11.0f;

         glPopMatrix();
      }

   // Next we will display some output to make things easier for us.  We will print
   // out the total amount of cubes that are in the viewing frustum (being rendered) and
   // the frames per second.
   glDisable(GL_LIGHTING);

      font.PrintText("Frustum Culling demo", -0.9f, 0.8f);
      font.PrintText("www.UltimateGameProgramming.com", -0.9f, 0.7f);

      sprintf(text, "Objects in frustum: %d", cubes);
      font.PrintText(text, -0.9f, 0.5f);

      int fps = GetFPS();
      sprintf(text, "Frames Per Second: %d", fps);
      font.PrintText(text, -0.9f, 0.3f);

   glEnable(GL_LIGHTING);


	// And thats all there is to it.
	SwapBuffers(g_HDC);                 // Display the new frame.
}


int GetFPS()
{
   // This will hold the number of frames per second in between seconds.
   static int FPS          = 0;
   static int CurrentFps   = 0;

   // This will be used to check if a second has passed.
   float nextSecond        = 0.0f;

   // This is the last second that occured.
   static float prevSecond	= 0.0f;

   // Add to the frames per second every time this function is called.
   CurrentFps++;

   // Get the second in millisecond then multiply it by 0.001 to convert this to
   // seconds.
   nextSecond = GetTickCount() * 0.001f;				

   // If the time we have now substracted from the previous second is greater than
   // or equal to 1 (i.e. if a second has passed) then we display the FPS number.
   if( nextSecond - prevSecond > 1.0f )
      {
         // Make the second we just got the previous second so we can used this the
         // next time the function is called.
	      prevSecond = nextSecond;

         FPS = CurrentFps;
		   
         // Reset the FPS counter.
         CurrentFps = 0;
      }

   return FPS;
}


void Shutdown()
{
   font.ClearGLFont();                 // Clear out the resources used by the font class.
}


// Recap:

// Frustum culling is one of the easiest ways to speed up your frame rate in a 3D scene.  By
// using frustum culling we culled out on average over 700 cubes from being rendered (on my
// machine anyway).  If we didn't use frustum culling we would have to render cubes we could
// not even see making our frame rate drop down.  So even if your scene is not
// all that complicated it would not hurt to use frustum culling.


// Copyright June 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com